#
# completion.rb
#
#   Copyright (c) 2001,2002 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

module ReFe

  class CompletionTable

    def initialize( cand = nil )
      @candidates = {}
      @table = {}
      if cand then
        cand.each do |c|
          add c
        end
      end
    end

    def inspect
      "#<#{type}>"
    end

    def add( cand )
      return if @candidates[cand]
      @candidates[cand] = true

      str = cand.downcase
      begin
        (@table[str] ||= []).push cand
        str.chop!
      end until str.empty?
    end

    def candidates
      @candidates.keys
    end

    def candidate?( name )
      @candidates.key? name
    end

    def complete( name )
      if @candidates[name] then
        [name]
      else
        (@table[name.downcase] || []).dup
      end
    end

    def complete_method( name )
      if @candidates[name] then
        return [name]
      end
      unless /[?!=]\z/ === name then
        return complete(name)
      end

      tail = name[-1]
      (@table[name.chop.downcase] || []).find_all {|m| m[-1] == tail }
    end
  
  end   # class CompletionTable


  class FCompletionTable

    def initialize( cand = nil )
      @tree = {}
      if cand then
        cand.each do |c|
          add c
        end
      end
    end

    def inspect
      "#<#{type}>"
    end

    def add( cand )
      a = cand.split(/_/)
      ident = a.pop
      t = @tree
      a.each do |prefix|
        t = (t[prefix] ||= {})
      end
      (t['-ident'] ||= {})[ ident ] = true
    end

    def complete( name )
      _comp( @tree, name.split(/_/), '#' ).collect {|n| n.sub(/\A\#_/, '') }
    end

    private

    def _comp( table, keys, path )
      k = keys[0]
      keys = keys[1..-1]
      k or return collect_all_sub_candidates(table, path)

      if complete_match_at_tail? table, keys, k then
        return [path + '_' + k]
      end
      if complete_match? table, k then
        ret = _comp( table[k], keys, path + '_' + k )
        return ret unless ret.empty?
      end

      find_candidates_at_tail( table, keys, k, path ) +
      find_candidates_from_children( table, keys, k, path )
    end

    def complete_match_at_tail?( table, keys, k )
      keys.empty? and it = table['-ident'] and it[k]
    end

    def complete_match?( table, k )
      table.key? k
    end

    def find_candidates_at_tail( table, keys, k, path )
      if keys.empty? and it = table['-ident'] then
        complete_word(it, k).collect {|n| path + '_' + n }
      else
        []
      end
    end

    def find_candidates_from_children( table, keys, k, path )
      complete_word(table, k).collect {|n|
              _comp table[n], keys, path + '_' + n }.flatten
    end

    def complete_word( table, key )
      cands = []

      if /[A-Z]/ === key then
        find_all_by(table, :each_key) {|k| k.index(key) == 0 }
      else
        find_all_by(table, :each_key) {|k| k.downcase.index(key) == 0 }
      end
    end

    def find_all_by( recv, method, &block )
      ret = []
      recv.__send__(method) {|a| yield(a) and ret.push a }
      ret
    end

    def collect_all_sub_candidates( table, path )
      collect_recursive(table).collect {|n| path + '_' + n }
    end

    def collect_recursive( table )
      table.collect {|k,t|
          if k == '-ident' then
            t.keys
          else
            collect_recursive(t).collect {|n| k + '_' + n }
          end
      }.flatten
    end
  
  end

end
