#
# database.rb
#
#   Copyright (c) 2001,2002 Minero Aoki <aamine@loveruby.net>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'refe/info'
require 'refe/completion'
require 'refe/mtable'
begin
  require 'refe/config'
rescue LoadError
end


module ReFe

  class SearchError < StandardError; end
  class ReFeFatalError < Exception; end


  class << self
    def database_dir( desig )
      desig                  and return desig
      ENV['REFE_DATA_DIR']   and return ENV['REFE_DATA_DIR']
      defined? REFE_DATA_DIR and return REFE_DATA_DIR

      raise ArgumentError, 'FATAL: refe database directory not given'
    end
  end


  class Database

    def initialize( dir, ronly, init )
      @dir = mustdir(dir)
      @readonly = ronly
      @init = @readonly ? false : init

      @m_table = nil
      @m_doc = nil
      @m_src = nil
      @mtof = nil
      @f_comptable = nil
      @f_src = nil
      @f_doc = nil

      @m_doc_updating = false
      @m_src_updating = false
      @f_doc_updating = false
      @f_src_updating = false
      @mtof_updating  = false
    end

    def mustdir( dir )
      File.directory? dir or raise ArgumentError, "is not directory: #{dir}"
      dir
    end

    def inspect
      "#<#{type}>"
    end

    #
    # register
    #

    def register_mdoc( name, str )
      m_doc_update
      @m_doc[ name ] = str
      @m_table.register_method name
    end

    def register_msrc( name, str )
      m_src_update
      @m_src[ name ] = str
      @m_table.register_method name
    end

    def register_class( name )
      m_table_update
      @m_table.register_class name
    end

    def register_mf_relation( m, f )
      mtof_update
      @mtof[m] = f
    end

    def register_fdoc( name, str )
      f_doc_update
      @f_doc[ name ] = str
      @f_comptable.add name
    end

    def register_fsrc( name, header, body, meta = nil )
      f_src_update
      @f_src[ name + '-header' ] = header
      @f_src[ name + '-body' ] = body
      @f_src[ name + '-meta' ] = meta if meta
      @f_comptable.add name
    end

    def self.index_updater( index, table )
      module_eval <<-"}}", __FILE__, __LINE__ + 1
        def #{index}_updating
          #{index}
          #{table}
          begin
            @#{index}_updating = true
            @#{index}.updating {
              yield
            }
          ensure
            save_#{table}
            @#{index}_updating = false
          end
        end

        def #{index}_update
          @#{index}_updating or
              raise 'index #{index} is not ready for update'
        end
        private :#{index}_update
      }}
    end

    index_updater :m_doc, :m_table
    index_updater :m_src, :m_table
    index_updater :f_doc, :f_comptable
    index_updater :f_src, :f_comptable

    alias class_updating m_doc_updating

    def m_table_update
      @m_src_updating or @m_doc_updating or
          raise 'mtable is not ready'
    end

    def mtof_updating
      mtof
      begin
        @mtof_updating = true
        yield
      ensure
        save_mtof
        @mtof_updating = false
      end
    end

    def mtof_update
      @mtof_updating or
          raise 'mtof is not ready for update'
    end

    #
    # lookup
    #

    def registered_classes
      m_table.registered_classes
    end

    def methods_of( c )
      m_table.methods_of(c)
    end

    def class_document( name )
      ''   # not implemented
    end

    def method_document( name )
      m_doc[name] or notfound 'method documentation', name
    end

    def method_definition( name )
      m_doc[name] or notfound 'method definition', name
    end

    def method_to_function( m )
      mtof[m] or raise SearchError, "function not found for method #{m}"
    end

    def function_document( name )
      f_doc[name] or notfound 'function document', name
    end

    def function_definition( name )
      File.basename( function_metadata(name) ) + "\n\n" +
      function_header(name) +
      function_body(name)
    end

    def function_header( name )
      f_src[name + '-header'] or notfound 'function header', name
    end

    def function_body( name )
      f_src[name + '-body'] or notfound 'function body', name
    end

    def function_metadata( name )
      f_src[name + '-meta'] or notfound 'function meta data', name
    end

    #
    # completion
    #

    def complete_class( c, t )
      m_table.complete_class c, t
    end

    def complete_method( c, t, m )
      m_table.complete_method c, t, m
    end

    def complete_method2( m )
      m_table.complete_method2 m
    end

    def complete_function( name )
      cand = f_comptable.complete(name)
      cand.empty? and compfail 'function', name
      cand
    end

    private

    def notfound( type, name )
      raise SearchError, "#{type} '#{name}' not registered"
    end

    def compfail( type, name )
      raise SearchError, "cannot expand #{type}: #{name}"
    end

    #
    # load
    #

    private

    %w( m_src m_doc f_src f_doc ).each do |name|
      module_eval( <<-End, __FILE__, __LINE__ + 1 )
          def #{name}
            @#{name} ||= loadobj('#{name.gsub(/_/, '')}', Index)
          end
      End
    end

    def m_table
      @m_table ||= loadobj(ID_MTABLE, TableLoader.new(MethodTable))
    end

    def mtof
      @mtof ||= loadobj(ID_MTOF, TableLoader.new(Hash))
    end

    def f_comptable
      @f_comptable ||= loadobj(ID_FCOMP, TableLoader.new(FCompletionTable))
    end

    def loadobj( ident, ld )
      path = @dir + '/' + ident

      if @init then
        ld.create_new(path)
      else
        begin
          ld.load(path)
        rescue Errno::ENOENT, IOError => err
          @readonly and raise ArgumentError,
                "ReFe database not initialized on #{@dir}: mkrefe first"
          ld.create_new(path)
        rescue StandardError => err
          raise ReFeFatalError,
              "unexpected error on database open: #{err.message} (#{err.type})"
        end
      end
    end


    def save_m_table
      dump ID_MTABLE, @m_table
    end

    def save_mtof
      dump ID_MTOF, @mtof
    end

    def save_f_comptable
      dump ID_FCOMP, @f_comptable
    end

    ID_MTABLE = 'mtable'
    ID_MTOF   = 'mtof'
    ID_FCOMP  = 'ftable'

    def dump( ident, obj )
      File.open( @dir + '/' + ident, 'w' ) {|f|
        Marshal.dump obj, f
      }
    end

  end


  # adapter class (file private)

  class TableLoader

    def initialize( c )
      @class = c
    end

    def create_new( fname )
      File.unlink fname if FileTest.file? fname
      @class.new
    end

    def load( fname )
      File.open( fname ) {|f|
        return Marshal.load(f)
      }
    end

  end


  class Index

    unless defined? DBM_CLASS then
      require 'gdbm'
      DBM_CLASS = GDBM
    end


    def self.create_new( fname )
      File.unlink fname         if FileTest.file? fname
      File.unlink fname + '.db' if FileTest.file? fname + '.db'
      new( fname )
    end

    def self.load( fname )
      File.file? fname         or
      File.file? fname + '.db' or raise IOError, "dbm file '#{fname}' not exist"
      new( fname )
    end


    def initialize( fname )
      @filename = fname
      @conn = nil
    end

    def dbm_class
      DBM_CLASS
    end

    def updating
      if @conn then
        yield
        return
      end
      dbm_class.open( @filename ) {|c|
        begin
          @conn = c
          yield
        ensure
          @conn = nil
        end
      }
    end

    def []( key )
      if @conn then
        @conn[key]
      else
        dbm_class.open(@filename) {|db|
          ret =db[key]
        }
      end
    end

    def []=( key, val )
      @conn or raise IOError, 'call #updating before []='
      @conn[key] = val
    end

    def size
      if @conn then
        @conn.size
      else
        dbm_class.open( @filename ) {|db|
          return db.size
        }
      end
    end

  end

end
